/* The main program */

#include <stdio.h>
#include <stdlib.h>
#include "common.h"
#include "kernels.h"
#include "gdrapi.h"
#include <pcilib.h>
#include <pcilib/kmem.h>
#include <pcilib/bar.h>
#include "ipedma.h"
#include <unistd.h>

#define KMEM_DEFAULT_FLAGS      (pcilib_kmem_flags_t)(PCILIB_KMEM_FLAG_HARDWARE | PCILIB_KMEM_FLAG_PERSISTENT | PCILIB_KMEM_FLAG_EXCLUSIVE)

#define KMEM_USE_RING           PCILIB_KMEM_USE(PCILIB_KMEM_USE_USER, 1)
#define KMEM_USE_DEFAULT        PCILIB_KMEM_USE(PCILIB_KMEM_USE_USER, 2)



int main()
{
    /* Initialisation of the APIs */
    assert_cu( cuInit(0) );
    gdr_t g = gdr_open();

    
    /* Manage NVIDIA GPU */
    printf("\nInitialisation of the GPU\n");
    CUdevice GPU;
    CUdevprop GPUProp;
    assert_cu( cuDeviceGet(&GPU,0) );
    assert_cu( cuDeviceGetProperties(&GPUProp,GPU) );
    char gpu_name[30] = {0};
    assert_cu (cuDeviceGetName (gpu_name, 30, GPU));
    printf("GPU: %s\n", gpu_name);
    
    /* Check context */
    CUcontext cuCtx;
    assert_cu( cuCtxCreate(&cuCtx,CU_CTX_MAP_HOST|CU_CTX_SCHED_AUTO,GPU) );
    assert_cu( cuCtxSetCurrent(cuCtx) );

    
    /* Try some stuff... */
    printf("Using binary data to feed FPGA...\n");
    char* data=(char*)calloc(4096,sizeof(*data));
    memset(data,0xAB,4096);
    init_to_send(data,4096);
    /* uint64_t nb=0xA10000003; */
    /* init_to_send(&nb,sizeof(nb)); */
    system("/home/mathiasb/sources/gpuFirstComm/launch.sh");
    
    
    /* Allocate memory on the device, pin and map */
    uint8_t flagValueToSet = 1;
    
    printf("\nMemory mapping with the GPU for pages\n");
    CUdeviceptr gpuPagePtr;
    assert_cu( cuMemAlloc(&gpuPagePtr,PAGE_SIZE) );
    assert_cu( cuPointerSetAttribute(&flagValueToSet,CU_POINTER_ATTRIBUTE_SYNC_MEMOPS,gpuPagePtr) );
    gdr_mh_t GPUMemHandlePage;
    assert_gdr( gdr_pin_buffer(g,gpuPagePtr,PAGE_SIZE,0,0,&GPUMemHandlePage) );
    void* gpuPageVa;
    assert_gdr( gdr_map(g,GPUMemHandlePage,&gpuPageVa,PAGE_SIZE) );
    gdr_info_t pageInfo;
    assert_gdr( gdr_get_info(g,GPUMemHandlePage,&pageInfo) );
    printf("gpuPagePtr = %llx\nBus ptr = %lx\nVA = %lx\nSize = %lu\n",gpuPagePtr,pageInfo.bus_addr,pageInfo.va,pageInfo.mapped_size);

    
    printf("Memory mapping with the GPU for descriptors\n");
    CUdeviceptr gpuDescPtr;
    assert_cu( cuMemAlloc(&gpuDescPtr,GPU_PAGE) );       
    assert_cu( cuPointerSetAttribute(&flagValueToSet,CU_POINTER_ATTRIBUTE_SYNC_MEMOPS,gpuDescPtr) );
    gdr_mh_t GPUMemHandleDesc;
    assert_gdr( gdr_pin_buffer(g,gpuDescPtr,GPU_PAGE,0,0,&GPUMemHandleDesc) );
    void* gpuDescVa;
    assert_gdr( gdr_map(g,GPUMemHandleDesc,&gpuDescVa,GPU_PAGE) );
    gdr_info_t descInfo;
    assert_gdr( gdr_get_info(g,GPUMemHandleDesc,&descInfo) );
    printf("gpuDescPtr = %llx\nBus ptr = %lx\nVA = %lx\nSize = %lu\n",gpuDescPtr,descInfo.bus_addr,descInfo.va,descInfo.mapped_size);    
    
    printf("All set\n");

    
    /* PCI */
    printf("\nSetting up the PCI\n");
    pcilib_t* pciCtx;
    char* pciVa;
    pciCtx = pcilib_open("/dev/fpga0",PCILIB_MODEL_DETECT);
    if( pciCtx == NULL )
    {
	printf("Cannot open a context for pci\n");
	exit( EXIT_FAILURE );
    }
    pciVa = pcilib_resolve_bar_address(pciCtx,0, 0);
    if( pciVa == NULL )
    {
	printf("Cannot resolve PCI physical adress to virtual\n");
	exit( EXIT_FAILURE );
    }

    CUdeviceptr dBAR;
    assert_cu( cuMemHostRegister((void*)pciVa,128,CU_MEMHOSTREGISTER_IOMEMORY) );
    assert_cu( cuMemHostGetDevicePointer(&dBAR,(void*)pciVa, 0) );

    
    /* Config PCI for Pages*/
    pcilib_kmem_handle_t* pciHandlePage;
    pciHandlePage = pcilib_alloc_kernel_memory(pciCtx, PCILIB_KMEM_TYPE_DMA_C2S_PAGE, 1, ((PAGE_SIZE%4096)?(4096 * (1 + PAGE_SIZE/4096)):PAGE_SIZE), 4096, KMEM_USE_DEFAULT, KMEM_DEFAULT_FLAGS);
    if( pciHandlePage == NULL )
    {
	printf("Cannot allocate  PCI kernel memory\n");
	exit( EXIT_FAILURE );
    }
    volatile void* pciMemPtrPage;
    uintptr_t pciBusPage;
    pciMemPtrPage = (uint64_t*) pcilib_kmem_get_block_ua(pciCtx,pciHandlePage,0);
    if( pciMemPtrPage == NULL )
    {
	printf("Cannot get PCI pointer to kernel memory\n");
	exit( EXIT_FAILURE );
    }
    pciBusPage = pcilib_kmem_get_block_ba(pciCtx,pciHandlePage,0);
    if( pciBusPage == 0 )
    {
	printf("Cannot get PCI Bus address on kernel memory\n");
	exit( EXIT_FAILURE );
    }

    /* Config PCI for Desc */
    pcilib_kmem_handle_t* pciHandleDesc;
    pciHandleDesc = pcilib_alloc_kernel_memory(pciCtx,PCILIB_KMEM_TYPE_CONSISTENT, 1, 128, 4096,KMEM_USE_RING, KMEM_DEFAULT_FLAGS);
    if( pciHandleDesc == NULL )
    {
	printf("Cannot allocate  PCI kernel memory\n");
	exit( EXIT_FAILURE );
    }
    volatile void* pciMemPtrDesc;
    uintptr_t pciBusDesc;
    pciMemPtrDesc = (uint64_t*) pcilib_kmem_get_block_ua(pciCtx,pciHandleDesc,0);
    if( pciMemPtrDesc == NULL )
    {
	printf("Cannot get PCI pointer to kernel memory\n");
	exit( EXIT_FAILURE );
    }
    pciBusDesc = pcilib_kmem_get_block_ba(pciCtx,pciHandleDesc,0);
    if( pciBusDesc == 0 )
    {
	printf("Cannot get PCI Bus address on kernel memory\n");
	exit( EXIT_FAILURE );
    }
    printf("pciMemPtrPage = %lx\npciMemPtrDesc = %lx\n",(uint64_t)pciMemPtrPage,(uint64_t)pciMemPtrDesc);

    /* volatile uint64_t* trackAddr = (uint64_t*)((char*)pciMemPtrDesc+DESCRIPTOR_OFFSET+2*sizeof(uint32_t)); */
    /* printf("Data were written at %p\n",trackAddr); */
    /* printf("trackAddr = %lx\n",*trackAddr); */
    *(uint32_t*)gpuDescVa=0;
    printf("counter = %d\n",*(int*)gpuDescVa);

    
    /* FPGA */
    printf("\nWorking on the FPGA\n");
    WR32(REG_RESET_DMA, 1);
    usleep(100000);
    WR32(REG_RESET_DMA, 0);
    usleep(100000);
    WR32_sleep(REG_PERF_COUNTER,0);
    WR32_sleep(REG_NUM_PACKETS_PER_DESCRIPTOR,2); //16);
    WR32_sleep(REG_PACKET_LENGTH,0x80000 | 64); // added flag
    WR32_sleep(REG_TIMER_THRESHOLD, 0x1);
    WR32_sleep(REG_UPDATE_THRESHOLD, 0x1);
    WR64_sleep(REG_UPDATE_COUNTER,descInfo.bus_addr);
    WR64_sleep(REG_UPDATE_ADDRESS,pciBusDesc+DESCRIPTOR_OFFSET);
    WR32_sleep(REG_CONTROL,CONTROL_ENABLE_READ|CONTROL_SOURCE_RX_FIFO);
    WR32_sleep(REG_DMA,1);
    WR32_sleep(REG_INTERCONNECT, 0x232); //0x262);
    WR32_sleep(REG_COUNTER,0x1);
    usleep(100000);
    printf("counter = %lx\n",*(uint64_t*)gpuDescVa);
    WR32_sleep(REG_PERF_COUNTER,1);
    usleep(100000);
    
    WR64(REG_DESCRIPTOR_ADDRESS,pageInfo.bus_addr);
    double lath = 4. *RD32 (0x14)/ 1000;
    double lath20 = 4. * RD32 (0x20) / 1000;
    printf("lath = %lf\nlath20 = %lf\n",lath,lath20);
    usleep(50000);
    lath = 4. *RD32 (0x14)/ 1000;
    lath20 = 4. * RD32 (0x20) / 1000;
    printf("lath = %lf\nlath20 = %lf\n",lath,lath20);
    printf("counter = %lx\n",*(uint64_t*)gpuDescVa);

    /* printf("pageInfo.bus_addr = %lx\ndescInfo.bus_addr = %lx\npciBusPage = %lx\npciBusDesc = %lx\n",pageInfo.bus_addr,descInfo.bus_addr,pciBusPage,pciBusDesc); */
    /* printf("gpuDescVa = %lx\ngpuPageVa = %lx\n",(uint64_t)gpuDescVa,(uint64_t)gpuPageVa); */
    /* printf("pciVa = %p\npciMemPtrPage = %lx\npciMemPtrDesc = %lx\n",pciVa,(uint64_t)pciMemPtrPage,(uint64_t)pciMemPtrDesc); */
    /* printf("gpuPagePtr = %llx\ngpuDescPtr = %llx\n",gpuPagePtr,gpuDescPtr); */

    /* printf("Data were written at %p\n",trackAddr); */
    /* printf("trackAddr = %lx\n",*trackAddr); */
    
    /* add_three_global<<< 1,1 >>>(gpuPagePtr); */
    /* assert_cu( cuCtxSynchronize() ); */
    /* printf("Received : %lx\n",*(uint64_t*)gpuPageVa); */
    
    char* getBack=(char*)calloc(4096,sizeof(*getBack));
    memcpy(getBack,(const void*)gpuPageVa,4096);
    for(int i=0;i<512;i++)
    {
    	printf("%hhx",getBack[i]);
    }
    printf("\n");
    printf("counter = %lx\n",*(uint64_t*)gpuDescVa);
    free(getBack);
    
    /* Close everything */
    printf("\nClosing the connections\n");
    WR32(REG_COUNTER, 0);
    WR32(REG_DMA, 0);
    WR32(REG_RESET_DMA, 1);
    usleep (100000);
    WR32(REG_RESET_DMA, 0);
    usleep (100000);

    pcilib_free_kernel_memory(pciCtx,pciHandleDesc,PCILIB_KMEM_FLAG_FORCE);
    pcilib_free_kernel_memory(pciCtx,pciHandlePage,PCILIB_KMEM_FLAG_FORCE);
    assert_cu( cuMemHostUnregister((void*) pciVa) );
    pcilib_close(pciCtx);
    assert_gdr( gdr_unmap(g,GPUMemHandlePage,gpuPageVa,PAGE_SIZE) );
    assert_gdr( gdr_unpin_buffer(g,GPUMemHandlePage) );
    assert_gdr( gdr_unmap(g,GPUMemHandleDesc,gpuDescVa,GPU_PAGE) );
    assert_gdr( gdr_unpin_buffer(g,GPUMemHandleDesc) );
    assert_gdr( gdr_close(g) );
    assert_cu( cuMemFree(gpuPagePtr) );
    assert_cu( cuMemFree(gpuDescPtr) );
    assert_cu( cuCtxDestroy(cuCtx) );
    
    printf("All Cleared\n");
    
    exit(EXIT_SUCCESS);
}
